<?php
define("WINDCAVE_VERSION", "1.0.1");

class windcaveEntity
{
    public $sessions_suffix =  "/v1/sessions";
    public $transactions_suffix =  "/v1/transactions";

    public $logger;
    public $type, $amount, $currency, $merchant_reference;

    public function windcaveEntity($logger, $type, $amount, $currency, $merchant_reference)
    {
        $this->logger = $logger;
        $this->type = $type;
        $this->amount = $amount;
        $this->currency = $currency;
        $this->merchant_reference = $merchant_reference;
    }
    
    public function sendPost($to_send, $XId, $suffix)
    {
        $payload = json_encode($to_send, JSON_UNESCAPED_UNICODE);

        $url = MODULE_PAYMENT_WINDCAVE_API_URL . $suffix;

        $this->logger->addLog('sending post request to ' . $url . '. contents: ' . str_replace("\n", " ", json_encode($to_send)));

        $authorization = base64_encode(MODULE_PAYMENT_WINDCAVE_API_USERNAME . ':' . MODULE_PAYMENT_WINDCAVE_API_KEY);

        $headers = array(
            'Content-Type:application/json; charset=utf-8',
            "Authorization: Basic " . $authorization,
            "X-ID: " . $XId,
            "X-Client-Version: osCommerce " . tep_get_version() . " ext " . WINDCAVE_VERSION
        );

        $ch = curl_init($url);
        curl_setopt($ch, CURLOPT_ENCODING, "");
        curl_setopt($ch, CURLOPT_POSTFIELDS, $payload);
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, FALSE);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, FALSE);

        $result = curl_exec($ch);

        $status_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);

        $response = json_decode($result);

        curl_close($ch);

        $this->logger->addLog('post response received. status code:' . $status_code . ' contents: ' . str_replace("\n", " ", $result));

        return array(
            'response' => $response,
            'status_code' => $status_code
        );
    }

    public function sendGet($id, $suffix)
    {
        $url = MODULE_PAYMENT_WINDCAVE_API_URL . $suffix . '/' . $id;
        //$url = '';

        $this->logger->addLog('sending get request to ' . $url);

        $ch = curl_init($url);

        $authorization = base64_encode(MODULE_PAYMENT_WINDCAVE_API_USERNAME . ':' . MODULE_PAYMENT_WINDCAVE_API_KEY);

        $headers = array(
            "Authorization: Basic " . $authorization
        );

        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, FALSE);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, FALSE);

        $result = curl_exec($ch);

        $response = json_decode($result);

        $status_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);

        $this->logger->addLog('get response received. status code: ' . $status_code);

        return array(
            'response' => $response,
            'status_code' => $status_code
        );
    }

    public function getXId()
    {
        $var1 = md5(uniqid(rand(), true));
        $var2 = md5(uniqid(rand(), true));
        return $var1 . $var2;
    }

    public function getLockFilePath()
    {
        return getcwd() . "/ext/modules/payment/windcave/locks";
    }

    public function getLockFileName()
    {
        return 'windcave_session_' . $this->session_id;
    }

    public function query($origin, $session_id, $id='')
    {
        if($origin == 'admin'){
            $entity = 'transaction';
        }else{
            $entity = 'session';
            $id = $session_id;
        }
        $this->logger->addLog('[' . $origin . '] starting query', $entity, $id);
        $response_result = $this->sendGet($session_id, $this->sessions_suffix);
        $response = $response_result['response'];
        $status_code = $response_result['status_code'];

        if ($status_code == 200) {
            $this->logger->addLog('[' . $origin . '] 200 response - determine result', $entity, $id);
                $last_transaction = $response->transactions[0];
                $approved = $last_transaction->authorised;
            if ($approved) {
                return array(
                    'result' => "success",
                    'txn_info' => $last_transaction
                );
            } else {
                return array(
                    'result' => "failure",
                    'txn_info' => $last_transaction
                );
            }
        } elseif ($status_code == 202) {
            $this->logger->addLog('[' . $origin . '] 202 response - result indeterminate, retry required', $entity, $id);
            return array(
                'result' => "retry",
                'txn_info' => ""
            );
        } elseif (round($status_code / 100, 0) == 4) { //4xx
            $error_list = $response->errors;
            $message = "";
            foreach ($error_list as $error) {
                $message = $message . $error->message . ", ";
            }
            substr_replace($message, "", -2);
            $this->logger->addLog($status_code . ' response - query failed. detailed error info: ' . $message);

            return array(
                'result' => "failure",
                'txn_info' => ""
            );
        } else {
            $this->logger->addLog($status_code . ' response, retry required');
            return array(
                'result' => "retry",
                'txn_info' => ""
            );
        }
    }
}
