<?php
class windcaveSession extends windcaveEntity
{

    private $retry_limit = 5;

    //private $logger;
    private $methods, $card_id, $store_card, $stored_card_indicator, $customer, $order_id, $language;
    public $session_url, $session_id;

    //public function windcaveSession($logger, $set_properties, $type = '', $amount = '', $currency = '', $merchant_reference = '', $methods = '', $card_id = '', $store_card = '', $stored_card_indicator = '', $customer = '', $order_id = '', $order_totals = '')
    public function windcaveSession($logger, $set_properties, $info = array())
    {
        parent::windcaveEntity($logger, $info['type'], $info['amount'], $info['currency'], $info['merchant_reference']);
        if ($set_properties) {
            $this->methods = $info['methods'];
            $this->card_id = $info['card_id'];
            $this->store_card = $info['store_card'];
            $this->stored_card_indicator = $info['stored_card_indicator'];
            $this->customer = $info['customer'];
            $this->order_id = $info['order_id'];
            $this->language = $info['language'];
        }
    }

    private function restoreCart()
    {
        global $cart;

        $stored_cart_data = tep_db_fetch_array(tep_db_query("select * from windcave_stored_cart where session_id = '" . $this->session_id . "' limit 1"))['contents'];
        $stored_cart = json_decode($stored_cart_data);
        $contents = json_decode(json_encode($stored_cart->contents), true);
        $cart = new shoppingCart;
        $cart->contents = $contents;
        $cart->total = $stored_cart->total;
        $cart->weight = $stored_cart->weight;
        $cart->cartID = $stored_cart->cartID;
        $cart->content_type = $stored_cart->content_type;

        return $cart;
    }

    private function findHPPLink($links)
    {
        foreach ($links as $link) {
            if ($link->rel == "hpp")
                return $link->href;
        }
        return "";
    }

    public function createSession($XId = "", $retries = 0)
    {
        global $cart, $customer_id;

        $retries = isset($_GET['retries']) ? $_GET['retries'] : 0;

        $this->logger->addLog('preparing to create session');

        $cust_id = "not logged";
        if (tep_session_is_registered('customer_id')) {
            $cust_id = $customer_id;
        }

        if ($XId == "") {
            $XId = $this->getXId();
        }

        $to_send = array(
            'type'                => $this->type,
            'amount'              => $this->amount,
            'currency'            => $this->currency,
            'merchantReference'   => $this->merchant_reference,
            'methods'             => $this->methods,
            'callbackUrls'        => array(
                'approved'        => tep_href_link('ext/modules/payment/windcave/callback.php', '', 'SSL'),
                'declined'        => tep_href_link('ext/modules/payment/windcave/callback.php', '', 'SSL'),
                'cancelled'       => tep_href_link('ext/modules/payment/windcave/callback.php', '', 'SSL'),
            ),
            'notificationUrl'     => tep_href_link('ext/modules/payment/windcave/notification.php', '', 'SSL'),
            'customer'            => $this->customer,
            'storeCard'           => $this->store_card,
            'storedCardIndicator' => $this->stored_card_indicator,
            'storeCustomer'       => false,
            'language'            => $this->language
        );

        if (defined('MODULE_PAYMENT_WINDCAVE_SESSION_TIMEOUT') && !empty(MODULE_PAYMENT_WINDCAVE_SESSION_TIMEOUT)) {
            $expires = MODULE_PAYMENT_WINDCAVE_SESSION_TIMEOUT * 60 * 60;
            $to_send['expires'] = date(DATE_RFC3339, time() + $expires);
        }

        if ($this->card_id != '') {
            $to_send['cardId'] = $this->card_id;
        }

        if ($this->stored_card_indicator != false) {
            $to_send['storedCardIndicator'] = $this->stored_card_indicator;
        }

        $response_result = $this->sendPost($to_send, $XId, $this->sessions_suffix);

        $response = $response_result['response'];
        $status_code = $response_result['status_code'];


        if ($status_code == 200) {
            $this->session_id = $response->sessionId;
            $this->logger->addLog('200 response - duplicate X-ID', 'session', $this->session_id);
            $last_transaction = $response->transactions[0];
            $approved = $last_transaction->authorised;
            $query = tep_db_query("select * from windcave_session where session_id = '" . $this->session_id . "'");
            if ($query->num_rows > 0) {
                $session = tep_db_fetch_array($query);
                if ($approved) {
                    $session['session_state'] = 'approved';
                } else {
                    $session['session_state'] = 'declined';
                }
                tep_db_perform('windcave_session', $session);
            }
            return array(
                'success' => true,
                'session_id' => $this->session_id
            );
        } elseif ($status_code == 202) { //session created
            $this->session_url = $this->findHPPLink($response->links);
            $this->session_id = $response->id;

            $this->logger->addLog('202 response - session created successfully. id: ' . $this->session_id);

            $order = new order();
            $items = array();

            foreach ($order->products as $product) {
                $items[] = array(
                    'id' => $product['id'],
                    'qty' => $product['qty']
                );
            }

            $stored_items = json_encode($items);

            $txn_data = array(
                'transaction_type' => $this->type,
                'amount' => $this->amount,
                'currency' => $this->currency,
                'session_id' => $this->session_id,
                'status' => PROCESSING_STATUS_WINDCAVE,
                'customer_id' => $cust_id,
                'hpp_link' => $this->session_url,
                'refund_status' => NOT_REFUNDED_STATUS_WINDCAVE,
                'void_status' => NOT_VOIDED_STATUS_WINDCAVE,
                'date_added' => 'now()',
                'order_id' => $this->order_id,
                'items' => $stored_items,
                'outstanding' => true,
            );

            if ($this->type == "purchase") {
                $txn_data["auth_complete_status"] = NOT_APPLICABLE_STATUS_AUTH_WINDCAVE;
                $txn_data["void_status"] = NOT_APPLICABLE_VOID_STATUS_WINDCAVE;
            } else {
                $txn_data["auth_complete_status"] = NOT_COMPLETED_STATUS_AUTH_WINDCAVE;
                $txn_data["refund_status"] = NOT_APPLICABLE_REFUND_STATUS_WINDCAVE;
            }

            if ($this->store_card) {
                $txn_data['store_card'] = 1;
            } else {
                $txn_data['store_card'] = 0;
            }

            tep_db_perform('windcave_transaction', $txn_data);
            $insert_id = tep_db_insert_id();

            $session_item_data = array(
                'session_id' => $this->session_id,
                'session_state' => 'pending',
                'windcave_id' => $insert_id
            );

            tep_db_perform('windcave_session', $session_item_data);

            $cart_store = json_encode($cart);

            $stored_cart_data = array(
                'session_id' => $this->session_id,
                'contents' => $cart_store
            );

            tep_db_perform('windcave_stored_cart', $stored_cart_data);

            $cart->reset(true);

            //tep_redirect($this->session_url);
            return array(
                'success' => true,
                'url' => $this->session_url,
                'session_id' => $this->session_id
            );
        } elseif (round($status_code / 100, 0) == 4) { //4xx
            $error_list = $response->errors;
            $message = "";
            foreach ($error_list as $error) {
                $message = $message . $error->message . ", ";
            }
            substr_replace($message, "", -2);
            $this->logger->addLog($status_code . ' response - session creation failed. detailed error info: ' . $message);
            return array(
                'success' => false
            );
        } else {
            $this->logger->addLog($status_code . ' response, retry required');
            $success =  $this->doRetry($retries, $XId);
            return array(
                'success' => $success
            );
        }
    }

    public function doRetry($retries, $XId)
    {
        if ($retries == null) {
            $retries = 0;
        }
        if ($retries > $this->retry_limit) {
            $this->logger->addLog('failure - gave up after ' . ($retries - 1) . ' retries');
            //tep_redirect(tep_href_link(FILENAME_CHECKOUT_SHIPPING, 'error_message=Payment failed', 'SSL'));
        } else {
            $this->logger->addLog('retrying - redirecting to wait script');
            tep_redirect(tep_href_link('ext/modules/payment/windcave/wait.php', 'XId=' . $XId . '&origin=checkout&retries=' . $retries, 'SSL'));
        }
    }

    public function callback()
    {
        global $cart, $currency;
        if (is_array($currency)) {
            $currency = $currency['code'];
        }
        $this->session_id = $_GET['sessionId'];
        $retries = isset($_GET['retries']) ? $_GET['retries'] : 0;

        if ($retries > $this->retry_limit) {
            $this->logger->addLog('[callback] failure - gave up after ' . ($retries - 1) . ' retries', 'session', $this->session_id);
            $this->restoreCart();
        }

        $this->logger->addLog('[callback] callback invoked', 'session', $this->session_id);

        $lockHandler = null;
        try {
            $lockHandler = new windcaveFileLock($this->getLockFileName(), $this->getLockFilePath());
            if (!$lockHandler->tryLock(false)) {
                // Retry here
                $this->logger->addLog('[callback] lock unavailable, backend notification is processing this. waiting until it finishes', 'session', $this->session_id);
                tep_redirect(tep_href_link('ext/modules/payment/windcave/wait.php', 'sessionId=' . $this->session_id . '&origin=callback&retries=' . $retries, 'SSL'));
            }

            $approved = false;
            $query = tep_db_query("select * from windcave_session where session_id = '" . $this->session_id . "'");
            if ($query->num_rows == 0) {
                $this->logger->addLog('[callback] session item does not exist for this session', 'session', $this->session_id);

                $transaction_check = tep_db_fetch_array(tep_db_query("select * from windcave_transaction where session_id = '" . $this->session_id . "'"));
                $approved = isset($transaction_check) && $transaction_check['status'] == APPROVED_STATUS_WINDCAVE;
            } else {
                $session = tep_db_fetch_array($query);
                $this->logger->addLog('[callback] session item found', 'session item', $session['session_id']);

                if ($session['session_state'] == 'processing') {
                    // This actually should not happen
                    $lockHandler->release();
                    $this->logger->addLog('[callback] processing state - backend notification is processing this. waiting until it finishes', 'session item', $session['session_id']);
                    tep_redirect(tep_href_link('ext/modules/payment/windcave/wait.php', 'sessionId=' . $this->session_id . '&origin=callback&retries=' . $retries, 'SSL'));
                }

                $this->logger->addLog('[callback] session state is "pending", attempting to process', 'session item', $session['session_id']);
                $session_response = $this->getQueryResult('callback');

                if ($session_response['result'] == 'retry') {
                    if ($retries < $this->retry_limit) {
                        $lockHandler->release();

                        $this->logger->addLog('[callback] retry required, redirecting to wait script', 'session', $this->session_id);
                        tep_redirect(tep_href_link('ext/modules/payment/windcave/wait.php', 'sessionId=' . $this->session_id . '&origin=callback&retries=' . $retries, 'SSL'));
                    } else {
                        $this->logger->addLog('[callback] failure - gave up after ' . $retries . ' retries', 'session', $this->session_id);
                    }
                }

                $this->processTransaction($session, $session_response['result'], 'callback', $session_response['txn_info']);
                $approved = $session_response['result'] == 'success';
            }

            $this->logger->addLog('[callback] releasing lock', 'session', $this->session_id);
            $lockHandler->release();
            if ($approved) {
                $this->logger->addLog('[callback] transaction approved', 'session', $this->session_id);
                tep_redirect(tep_href_link('ext/modules/payment/windcave/blank_page.php', 'approved=true', 'SSL'));
            } else {
                $this->logger->addLog('[callback] transaction declined', 'session', $this->session_id);
                $cart =  $this->restoreCart();
                tep_redirect(tep_href_link('ext/modules/payment/windcave/blank_page.php', 'approved=false', 'SSL'));
            }
        } catch(\Exception $e) {
            $this->logger->addLog('[callback] exception on locking the file. ' . $e->getMessage() . $e->getTraceAsString(), 'session', $this->session_id);
            if (isset($lockHandler)) {
                $lockHandler->release();
            }
            tep_redirect(tep_href_link('ext/modules/payment/windcave/blank_page.php', 'approved=false', 'SSL'));
        }
    }

    public function notification()
    {
        $this->session_id = $_GET['sessionId'];

        if ($this->session_id == null) {
            $response = json_decode(file_get_contents('php://input'));
            $this->logger->addLog('[notification]' . str_replace("\n", ' ', file_get_contents('php://input')));
            $this->session_id = $response->sessionId;
        }

        $this->logger->addLog('[notification] notification invoked', 'session', $this->session_id);


        $lockHandler = null;
        try {
            $lockHandler = new windcaveFileLock($this->getLockFileName(), $this->getLockFilePath());
            if (!$lockHandler->tryLock(false)) {
                // Retry here
                $this->logger->addLog('[notification] lock unavailable, callback is processing this. waiting until it finishes', 'session', $this->session_id);
                http_response_code(500);
                return;
            }

            $query = tep_db_query("select * from windcave_session where session_id = '" . $this->session_id . "'");
            if ($query->num_rows > 0) {
                $session = tep_db_fetch_array($query);
                $this->logger->addLog('[notification] session item found', 'session item', $session['session_id']);
            } else {
                $this->logger->addLog('[notification] session item does not exist for this session. Return 200 and give up.', 'session', $this->session_id);
                http_response_code(200);
                $lockHandler->release();
                return;
            }

            if ($session['session_state'] == 'pending') {
                $this->logger->addLog('[notification] session state is "pending", attempting to process', 'session item', $session['session_id']);
                $session_response = $this->query('notification', $this->session_id);
                if ($session_response['result'] == 'retry') {
                    $this->logger->addLog('[notification]retry required - sending 500 response', 'session', $this->session_id);
                    http_response_code(500);
                    $lockHandler->release();
                    return;
                }

                $this->processTransaction($session, $session_response['result'], 'notification', $session_response['txn_info']);

                $this->logger->addLog('[notification] releasing lock', 'session', $this->session_id);
            }
            $lockHandler->release();
        } catch (\Exception $e) {
            $this->logger->addLog('[notification] exception on locking the file. ' . $e->getMessage() . $e->getTraceAsString(), 'session', $this->session_id);
            if (isset($lockHandler)) {
                $lockHandler->release();
            }
        }
        http_response_code(200);
        return;
    }

    private function getQueryResult($origin)
    {
        $this->logger->addLog('[' . $origin . '] attempting to query for session result', 'session', $this->session_id);
        $session_response = $this->query($origin, $this->session_id);
        return $session_response;
    }

    private function processTransaction($session, $result, $origin, $txn_info = false)
    {
        $this->logger->addLog('[' . $origin . '] re-checking session item state', 'session', $this->session_id);
        if ($session['session_state'] == 'pending') {
            $this->logger->addLog('[' . $origin . '] pending state as expected', 'session', $this->session_id);
            tep_db_query("update windcave_session set session_state = 'processing' where session_id = '" . $this->session_id . "'");
            if ($result == 'success') {
                $this->updateTransaction($txn_info, true, $session, $origin);
            } else {
                $this->updateTransaction($txn_info, false, $session, $origin);
            }
            tep_db_query("delete from windcave_session where session_id = '" . $this->session_id . "'");
        } else {
            $this->logger->addLog('[' . $origin . '] not pending state - error: state changed between checking for lock and locking', 'session', $this->session_id);
        }
    }

    private function updateTransaction($txn_info, $approved, $session, $origin)
    {
        $comments = 'This transaction was declined by Windcave';
        $this->logger->addLog('[' . $origin . '] updating transaction for this session', 'session', $this->session_id);
        if ($txn_info != false) {
            $data = array(
                "response_code"          => $txn_info->reCo,
                "response_text"          => $txn_info->responseText,
                "payment_method"         => $txn_info->method,
                "payment_authentication" => $txn_info->liabilityIndicator,
                "avs_result"             => $txn_info->avs->avsResultDescription,
                "avs_action"             => $txn_info->avs->avsActionName,
                "cvc2_result"            => $txn_info->cvc2ResultCode,
                "session_id"             => $this->session_id,
                "transaction_id"         => $txn_info->id
            );
            $transaction_id = $txn_info->id;

            if ($txn_info->card != null) {
                $card_info = $txn_info->card;
                if ($card_info->cardNumber != null) {
                    $data["card_mask"] = $txn_info->card->cardNumber;
                } else {
                    $data["card_mask"] = "000000........00";
                }

                if ($card_info->dateExpiryMonth != null) {
                    $data["expiry_date_month"] = $txn_info->card->dateExpiryMonth;
                } else {
                    $data["expiry_date_month"] = "";
                }

                if ($card_info->dateExpiryYear != null) {
                    $data["expiry_date_year"] = $txn_info->card->dateExpiryYear;
                } else {
                    $data["expiry_date_year"] = "";
                }

                if ($card_info->type != null) {
                    $data["card_type"] = $txn_info->card->type;
                } else {
                    $data["card_type"] = "unknown";
                }

                // Google Pay does not return the cardholder name
                if ($txn_info->card->cardHolderName != null) {
                    $data["cardholder_name"] = $txn_info->card->cardHolderName;
                } else {
                    $data["cardholder_name"] = "";
                }
            }

            if ($txn_info->authCode != null) {
                $data["auth_code"] = $txn_info->authCode;
            } else {
                $data["auth_code"] = "";
            }
            if ($txn_info->risk != null) {
                $data["risk_action"] = $txn_info->risk->action;
                $data["risk_summary_text"] = $txn_info->risk->summaryText;
            } else {
                $data["risk_action"] = "";
                $data["risk_summary_text"] = "";
            }
            if ($txn_info->card->id != null) {
                $data['card_id'] = $txn_info->card->id;
            } else {
                $data['card_id'] = "";
            }
            if ($approved) {
                $this->logger->addLog('[' . $origin . '] transaction approved', 'tranaction', $transaction_id);
                $data['status'] = APPROVED_STATUS_WINDCAVE;
                if ($txn_info->type == 'auth') {
                    $data["auth_complete_status"] = NOT_COMPLETED_STATUS_AUTH_WINDCAVE;
                    $data['void_status'] = NOT_VOIDED_STATUS_WINDCAVE;
                } else {
                    $data["auth_complete_status"] = NOT_APPLICABLE_STATUS_AUTH_WINDCAVE;
                    $data['refund_status'] = NOT_REFUNDED_STATUS_WINDCAVE;
                }
            } else {
                $this->logger->addLog('[' . $origin . '] transaction declined', 'tranaction', $transaction_id);
                $data['status'] = DECLINED_STATUS_WINDCAVE;
                $data['refund_status'] =  NOT_APPLICABLE_REFUND_STATUS_WINDCAVE;
                $data['void_status'] =  NOT_APPLICABLE_VOID_STATUS_WINDCAVE;
                $data["auth_complete_status"] = NOT_APPLICABLE_STATUS_AUTH_WINDCAVE;
            }
            $transaction = tep_db_fetch_array(tep_db_query("select * from windcave_transaction where session_id = '" . $this->session_id . "'"));

            if ($approved) {
                tep_db_query("delete from windcave_stored_cart where session_id = '" . $this->session_id . "'");
                $comments = 'This transaction was approved by Windcave';
            }

            tep_db_perform('windcave_transaction', $data, 'update', "session_id = '" . $this->session_id . "'");
        } else {
            $approved  = false;
            $status = DECLINED_STATUS_WINDCAVE;
            tep_db_query("update windcave_transaction set status = '" . $status . "' where session_id = '" . $this->session_id . "'");
        }

        $query = tep_db_query("select * from windcave_transaction where session_id = '" . $this->session_id . "' limit 1");
        $transaction = tep_db_fetch_array($query);
        if (($transaction['customer_id'] != 'not logged') && ($transaction['store_card'] == 1) && ($transaction['card_id'] != null)) {
            $this->logger->addLog('[' . $origin . '] ok to store card for this customer', 'tranaction', $transaction_id);
            $query = tep_db_query("select * from windcave_saved_card where card_id = '" . $txn_info->card->id . "' and customer_id = '" . $transaction['customer_id'] . "' limit 1");
            if (!($query->num_rows > 0)) {
                $card_data = array(
                    'customer_id'       => $transaction['customer_id'],
                    'card_id'           => $transaction['card_id'],
                    'card_mask'         => $transaction['card_mask'],
                    "expiry_date_month" => $transaction['expiry_date_month'],
                    "expiry_date_year"  => $transaction['expiry_date_year']
                );
                tep_db_perform('windcave_saved_card', $card_data);
            } else {
                $this->logger->addLog('[' . $origin . '] identical card already exists for this customer - dont store anything', 'tranaction', $transaction_id);
            }
        }

        if ($transaction['order_id'] != 0) {
            if ($approved) {
                $comments .= "\n" . 'For more information, please see Windcave Transactions page';
                $comments .= "\n" . 'Order ID: ' . $transaction['order_id'];
                $this->updateOrdersStatus($transaction['order_id'], MODULE_PAYMENT_WINDCAVE_TRANSACTION_ORDER_STATUS_ID, $comments, !$approved);
                $this->updateOrdersStatus($transaction['order_id'], MODULE_PAYMENT_WINDCAVE_ORDER_STATUS_ID);
            } else {
                $order_query = tep_db_query("select products_id, products_quantity from " . TABLE_ORDERS_PRODUCTS . " where orders_id = '" . (int) $transaction['order_id'] . "'");
                while ($order = tep_db_fetch_array($order_query)) {
                    tep_db_query("update " . TABLE_PRODUCTS . " set products_quantity = products_quantity + " . $order['products_quantity'] . ", products_ordered = products_ordered - " . $order['products_quantity'] . " where products_id = '" . (int) $order['products_id'] . "'");
                }
                tep_db_query("delete from " . TABLE_ORDERS . " where orders_id = '" . (int) $transaction['order_id'] . "'");
                tep_db_query("delete from " . TABLE_ORDERS_PRODUCTS . " where orders_id = '" . (int) $transaction['order_id'] . "'");
                tep_db_query("delete from " . TABLE_ORDERS_PRODUCTS_ATTRIBUTES . " where orders_id = '" . (int) $transaction['order_id'] . "'");
                tep_db_query("delete from " . TABLE_ORDERS_STATUS_HISTORY . " where orders_id = '" . (int) $transaction['order_id'] . "'");
                tep_db_query("delete from " . TABLE_ORDERS_TOTAL . " where orders_id = '" . (int) $transaction['order_id'] . "'");

                tep_db_query("update windcave_transaction set outstanding = '" . false . "' where transaction_id = '" . $transaction['transaction_id'] . "'");
            }
        }
    }
    private function updateOrdersStatus($order_id, $status, $comments = '', $update_table_orders = true)
    {
        if ($update_table_orders) {
            tep_db_query("update " . TABLE_ORDERS . " set orders_status = '" . $status . "', last_modified = now() where orders_id = '" . $order_id . "'");
        }

        $sql_data_array = array(
            'orders_id' => $order_id,
            'orders_status_id' => $status,
            'date_added' => 'now()',
            'customer_notified' => '0',
            'comments' => $comments
        );

        tep_db_perform(TABLE_ORDERS_STATUS_HISTORY, $sql_data_array);
    }
}
