<?php
class windcaveTransaction extends windcaveEntity
{

    private $parent_type, $parent_id;
    private $id;
    private $customer_id, $order_id;
    private $retry_limit = 5;
    private $statuses;

    //public function windcaveTransaction($logger, $type, $parent_type, $parent_id, $statuses = array(), $amount = '', $currency = '', $merchant_reference = '', $customer_id = '')
    public function windcaveTransaction($logger, $info)
    {
        parent::windcaveEntity($logger, $info['type'], $info['amount'], $info['currency'], $info['merchant_reference']);
        $this->parent_type = $info['parent_type'];
        $this->parent_id = $info['parent_id'];
        $this->customer_id = $info['customer_id'];
        $this->order_id = $info['merchant_reference'];
        $this->statuses = $info['statuses'];
    }

    public function startTransaction($target_types, $amount = 'none')
    {
        $transaction_success = false;
        $message = "";

        $XId = $this->getXId();
        $query = tep_db_query("select * from windcave_transaction where parent_transaction_id = '" . $this->parent_id . "'");
        if ($query->num_rows > 0) {
            for ($i = 0; $i < $query->num_rows; $i++) {
                $child_transaction = tep_db_fetch_array($query);
                if (($child_transaction['transaction_type'] == $this->type) && ($child_transaction['status'] != DECLINED_STATUS_WINDCAVE)) {
                    if ($this->type != 'refund') {
                        $this->logger->addLog("error - duplicate $this->type (existing $this->type ID: " . $child_transaction['transaction_id'] . "'", 'transaction', $this->parent_id);
                        return array(
                            'success' => $transaction_success,
                            'message' => "duplicate $this->type attempted"
                        );
                    }
                }
                if ($child_transaction['status'] == PROCESSING_STATUS_WINDCAVE) {
                    $this->logger->addLog("error - processing $this->type exists (ID: " . $child_transaction['transaction_id'] . ")" . 'transaction', $this->parent_id);
                    return array(
                        'success' => $transaction_success,
                        'message' => "processing $this->type transaction exists (ID: " . $child_transaction['transaction_id'] . ")"
                    );
                }
            }
        }
        if (!in_array($this->parent_type, $target_types)) {
            $this->logger->addLog("error - tried to $this->type unexpected transaction type - type was $this->parent_type", 'transaction', $this->parent_id);
            $message = "tried to $this->type unexpected transaction type - type was $this->parent_type";
        } else {
            $result = '';
            $this->updateStatus($this->statuses['parent_txn_processing']);
            if (true) { //fully credited
                $this->logger->addLog("attempting to $this->type this transaction", 'transaction', $this->parent_id);
                $result = $this->createTransaction($XId, 0);
                $success = $result['success'];
                $transaction_created = $result['transaction_created'];
            } else {
                //fully credited
            }
            if ($success) {
                $this->logger->addLog("$this->type successful");
                $this->updateStatus($this->statuses['parent_txn_complete']);
                //add order history?
                $transaction_success = true;
            } else {
                if (!$transaction_created) {
                    $this->updateStatus($this->statuses['parent_txn_previous']);
                } else {
                    $new_transaction = tep_db_fetch_array(tep_db_query("select * from windcave_transaction where transaction_id = '" . $result['transaction_id'] .  "' limit 1"));
                    if ($new_transaction['status'] == DECLINED_STATUS_WINDCAVE) {
                        $this->updateStatus($this->statuses['parent_txn_previous']);
                    }
                }
                $this->logger->addLog("$this->type unsuccessful", 'transaction', $this->parent_id);
            }
        }
        return array(
            'success' => $transaction_success,
            'message' => $message
        );
    }

    private function createTransaction($XId, $retries)
    {
        $original_transaction = tep_db_fetch_array(tep_db_query("select * from windcave_transaction where transaction_id = '" . $this->parent_id . "' limit 1"));
        if ($this->amount == '') {
            $this->amount = $original_transaction['amount'];
        }

        switch ($this->type) {
            case 'refund':
                $this->merchant_reference = 'Refund ' . $this->merchant_reference;
                break;
            case 'void':
                $this->merchant_reference = 'Void ' . $this->merchant_reference;
                break;
        }

        $to_send = array(
            'type' => $this->type,
            'transactionId' => $this->parent_id,
            'amount' => $this->amount,
            'currency' => $this->currency,
            'merchantReference' => $this->merchant_reference
        );

        $response_result = $this->sendPost($to_send, $XId, $this->transactions_suffix);
        $response = $response_result['response'];
        $status_code = $response_result['status_code'];

        $data = array(
            'order_id' => $this->order_id,
            'transaction_id' => '',
            'parent_transaction_id' => $this->parent_id,
            'transaction_type' => $this->type,
            'amount' => $this->amount,
            'currency' => $this->currency,
            'status' => PROCESSING_STATUS_WINDCAVE,
            'auth_complete_status' => NOT_APPLICABLE_STATUS_AUTH_WINDCAVE,
            'customer_id' => $this->customer_id,
            'fully_credited' => false,
            'hpp_link' => '',
            'session_id' => $original_transaction['session_id'],
            'store_card' => false,
            'refund_status' => NOT_REFUNDED_STATUS_WINDCAVE,
            'void_status' => NOT_APPLICABLE_VOID_STATUS_WINDCAVE,
            'outstanding' => true,
            'date_added' => 'now()',
        );
        if ($this->type != 'complete') {
            $data['refund_status'] = NOT_APPLICABLE_REFUND_STATUS_WINDCAVE;
        }

        if ($status_code == 200 || $status_code == 201) {
            $this->id = $response->id;
            if ($this->type == 'void') {
                $this->id = $this->id . '_void';
            }
            $status = DECLINED_STATUS_WINDCAVE;
            $success = false;
            if ($response->authorised) {
                $status = APPROVED_STATUS_WINDCAVE;
                $success = true;
            } else {
                $data['refund_status'] = NOT_APPLICABLE_REFUND_STATUS_WINDCAVE;
                $data['void_status'] = NOT_APPLICABLE_VOID_STATUS_WINDCAVE;
            }


            $data['transaction_id'] = $this->id;
            $data['status'] = $status;
            $this->logger->addLog($status_code . ' response, transaction created successfully');
            tep_db_perform('windcave_transaction', $data);
            $this->logger->addLog("created $this->type transaction for this transaction", 'transaction', $this->parent_id);
            return array(
                'success' => $success,
                'transaction_created' => true,
                'transaction_id' => $this->id
            );
        } elseif ($status_code == 202) {
            $this->id = $response->id;
            if ($this->type == 'void') {
                $this->id = $this->id . '_void';
            }
            $data['transaction_id'] = $this->id;
            $this->logger->addLog('202 response, result indeterminate');
            tep_db_perform('windcave_transaction', $data);
            $this->logger->addLog("created $this->type for this transaction", 'transaction', $this->parent_id);
            sleep(2);
            $this->logger->addLog('attempting to query transaction', 'transaction', $this->id);
            $result = $this->doQuery(0, $original_transaction['session_id'], $this->id);
            if ($result != 'indeterminate') {
                if ($result == 'success') {
                    tep_db_query("update windcave_transaction set status = '" . APPROVED_STATUS_WINDCAVE . "' where transaction_id = '" . $this->id . "'");
                }
                else{
                    $this->updateStatus($this->statuses['parent_txn_previous']);
                    tep_db_query("update windcave_transaction set status = '" . DECLINED_STATUS_WINDCAVE . "' where transaction_id = '" . $this->id . "'");
                }
            }
            return array(
                'success' => $result == 'success',
                'transaction_created' => true,
                'transaction_id' => $this->id
            );
        } elseif (round($status_code / 100, 0) == 4) {
            $error_list = $response->errors;
            $message = "";
            foreach ($error_list as $error) {
                $message = $message . $error->message . ", ";
            }
            substr_replace($message, "", -2);
            $this->logger->addLog($status_code . ' response - session creation failed. detailed error info: ' . $message);
            return array(
                'success' => false,
                'transaction_created' => false
            );
        } else {
            $this->logger->addLog("$status_code response, retry required");
            if ($retries < $this->retry_limit) {
                sleep(5);
                $this->logger->addLog("reattempting transaction creation");
                return $this->createTransaction($XId, ++$retries);
            } else {
                $this->logger->addLog("transaction creation failed, gave up after $retries retries");
                return array(
                    'success' => false,
                    'transaction_created' => false
                );
            }
        }
    }

    private function updateStatus($status)
    {
        switch ($this->type) {
            case 'refund':
                tep_db_query("update windcave_transaction set refund_status = '" . $status . "' where transaction_id = '" . $this->parent_id . "'");
                break;
            case 'complete':
                tep_db_query("update windcave_transaction set auth_complete_status = '" . $status . "' where transaction_id = '" . $this->parent_id . "'");
                break;
            case 'void':
                tep_db_query("update windcave_transaction set void_status = '" . $status . "' where transaction_id = '" . $this->parent_id . "'");
                break;
        }
    }

    public function doQuery($retries, $session_id, $query_transaction_id)
    {
        $result = $this->query('admin', $session_id, $query_transaction_id)['result'];
        if ($result == 'retry') {
            if ($retries < $this->retry_limit) {
                $this->logger->addLog('requerying');
                sleep(2);
                return $this->doQuery(++$retries, $session_id, $query_transaction_id);
            } else {
                $this->logger->addLog("query failure, gave up after $retries retries");
                return 'indeterminate';
            }
            //} elseif ($result == 'failure') {
            //    $this->updateStatus($this->statuses['parent_txn_previous']);
            //    tep_db_query("update windcave_transaction set status = '" . DECLINED_STATUS_WINDCAVE . "' where transaction_id ='" . $query_transaction_id . "'");
        }
        return $result;
    }
}
